# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.

"""
This package is an overarching package for modules performing various analyses
of materials data.
"""

